import os

import click
import pystac

from stactools.landsat.utils import transform_stac_to_stac
from stactools.landsat.stac import create_stac_item


def create_landsat_command(cli):
    """Creates a command group for working
    with Landsat metadata from USGS' Collection 2
    """
    @cli.group(
        'landsat',
        short_help=("Commands for working with Landsat Collection 2 metadata.")
    )
    def landsat():
        pass

    @landsat.command(
        "create-item",
        short_help="Create a STAC item from collection 2 scene metadata.")
    @click.option("--level",
                  type=click.Choice(['level-1', 'level-2'],
                                    case_sensitive=False),
                  default="level-2",
                  show_default=True,
                  help="Product level to process")
    @click.option("--mtl", required=True, help="HREF to an MTL file.")
    @click.option("--output",
                  required=True,
                  help="HREF of diretory in which to write the item.")
    def create_item_cmd(level: str, mtl: str, output: str):
        """Creates a STAC Item for a Landsat 8 C2 Level-2 scene's products.

        All asset paths are based on the MTL path, as all assets are assumed to
        reside in the same directory/blob prefix/etc.
        """
        if level != 'level-2':
            raise click.BadOptionUsage("level",
                                       "Only level-2 currently implemented.")

        item = create_stac_item(mtl_xml_href=mtl)
        item.set_self_href(os.path.join(output, f'{item.id}.json'))
        item.save_object()

    @landsat.command(
        "convert",
        short_help="Convert a USGS STAC 0.7 Item to an updated STAC Item")
    @click.option("--stac", "-s", required=True, help="Path to a STAC file.")
    @click.option(
        "--enable-proj",
        "-p",
        default=True,
        is_flag=True,
        help="Enable the proj extension. Requires access to blue band.")
    @click.option("--dst", "-d", help="Output directory")
    def convert_cmd(stac, enable_proj, dst):
        in_item = pystac.Item.from_file(stac)
        item = transform_stac_to_stac(in_item, enable_proj=enable_proj)

        item_path = os.path.join(dst, '{}.json'.format(item.id))
        item.set_self_href(item_path)
        item.save_object()

    return landsat



import datetime

import dateutil.parser
import rasterio
from pystac import Item, Link, MediaType, STACError
from rasterio import RasterioIOError
from shapely.geometry import box, mapping, shape


def _parse_date(in_date: str) -> datetime.datetime:
    """
    Try to parse a date and return it as a datetime object with no timezone
    """
    parsed_date = dateutil.parser.parse(in_date)
    return parsed_date.replace(tzinfo=datetime.timezone.utc)


def transform_mtl_to_stac(metadata: dict) -> Item:
    """
    Handle USGS MTL as a dict and return a STAC item.

    NOT IMPLEMENTED

    Issues include:
        - There's no reference to UTM Zone or any other CRS info in the MTL
        - There's no absolute file path or reference to a URI to find data.
    """
    LANDSAT_METADATA = metadata["LANDSAT_METADATA_FILE"]
    product = LANDSAT_METADATA["PRODUCT_CONTENTS"]
    projection = LANDSAT_METADATA["PROJECTION_ATTRIBUTES"]
    image = LANDSAT_METADATA["IMAGE_ATTRIBUTES"]
    proessing_record = LANDSAT_METADATA["LEVEL2_PROCESSING_RECORD"]

    scene_id = product["LANDSAT_PRODUCT_ID"]

    xmin, xmax = float(projection["CORNER_LL_LON_PRODUCT"]), float(
        projection["CORNER_UR_LON_PRODUCT"])
    ymin, ymax = float(projection["CORNER_LL_LAT_PRODUCT"]), float(
        projection["CORNER_UR_LAT_PRODUCT"])
    geom = mapping(box(xmin, ymin, xmax, ymax))
    bounds = shape(geom).bounds

    # Like: "2020-01-01" for date and  "23:08:52.6773140Z" for time
    acquired_date = _parse_date(
        f"{image['DATE_ACQUIRED']}T{image['SCENE_CENTER_TIME']}")
    created = _parse_date(proessing_record["DATE_PRODUCT_GENERATED"])

    item = Item(id=scene_id,
                geometry=geom,
                bbox=bounds,
                datetime=acquired_date,
                properties={})

    # Common metadata
    item.common_metadata.created = created
    item.common_metadata.platform = image["SPACECRAFT_ID"]
    item.common_metadata.instruments = [
        i.lower() for i in image["SENSOR_ID"].split("_")
    ]

    # TODO: implement these three extensions
    item.ext.enable("eo")
    item.ext.enable("view")
    item.ext.enable("projection")

    return item


def transform_stac_to_stac(item: Item,
                           enable_proj: bool = True,
                           self_link: str = None,
                           source_link: str = None) -> Item:
    """
    Handle a 0.7.0 item and convert it to a 1.0.0.beta2 item.
    If `enable_proj` is true, the assets' geotiff files must be accessible.
    """
    # Clear hierarchical links
    item.set_parent(None)
    item.set_root(None)

    # Remove USGS extension and add back eo
    item.ext.enable("eo")

    # Add and update links
    if self_link:
        item.links.append(Link(rel="self", target=self_link))
    if source_link:
        item.links.append(
            Link(rel="derived_from",
                 target=source_link,
                 media_type="application/json"))

    # Add some common fields
    item.common_metadata.constellation = "Landsat"
    item.common_metadata.instruments = [
        i.lower() for i in item.properties["eo:instrument"].split("_")
    ]
    del item.properties["eo:instrument"]

    # Handle view extension
    item.ext.enable("view")
    item.ext.view.off_nadir = item.properties["eo:off_nadir"]
    del item.properties["eo:off_nadir"]

    if enable_proj:
        # Enabled projection
        item.ext.enable("projection")

        obtained_shape = None
        obtained_transform = None
        crs = None
        for name, asset in item.assets.items():
            if "geotiff" in asset.media_type:
                # retrieve shape, transform and crs from the first geotiff file among the assets
                if not obtained_shape:
                    try:
                        with rasterio.open(asset.href) as opened_asset:
                            obtained_shape = opened_asset.shape
                            obtained_transform = opened_asset.transform
                            crs = opened_asset.crs.to_epsg()
                            # Check to ensure that all information is present
                            if not obtained_shape or not obtained_transform or not crs:
                                raise STACError(
                                    f"Failed setting shape, transform and csr from {asset.href}"
                                )

                    except RasterioIOError as io_error:
                        raise STACError(
                            "Failed loading geotiff, so not handling proj fields"
                        ) from io_error

                    item.ext.projection.set_transform(obtained_transform,
                                                      asset=asset)
                    item.ext.projection.set_shape(obtained_shape, asset=asset)
                asset.media_type = MediaType.COG

        # Now we have the info, we can make the fields
        item.ext.projection.epsg = crs

    # Remove .TIF from asset names
    item.assets = {
        name.replace(".TIF", ""): asset
        for name, asset in item.assets.items()
    }

    return item


def stac_api_to_stac(uri: str) -> Item:
    """
    Takes in a URI and uses that to feed the STAC transform
    """

    return transform_stac_to_stac(item=Item.from_file(uri),
                                  source_link=uri,
                                  enable_proj=False)



from typing import Optional
import pystac

from stactools.core.io import ReadHrefModifier
from stactools.landsat.assets import (ANG_ASSET_DEF, COMMON_ASSET_DEFS,
                                      SR_ASSET_DEFS, THERMAL_ASSET_DEFS)
from stactools.landsat.constants import (L8_EXTENSION_SCHEMA, L8_INSTRUMENTS,
                                         L8_ITEM_DESCRIPTION, L8_PLATFORM)
from stactools.landsat.mtl_metadata import MtlMetadata
from stactools.landsat.ang_metadata import AngMetadata


def create_stac_item(
        mtl_xml_href: str,
        read_href_modifier: Optional[ReadHrefModifier] = None) -> pystac.Item:
    """Creates a Landsat 8 C2 L2 STAC Item.

    Reads data from a single scene of
    Landsat Collection 2 Level-2 Surface Reflectance Product data.

    Uses the MTL XML HREF as the bases for other files; assumes that all
    files are co-located in a directory or blob prefix.
    """
    base_href = '_'.join(mtl_xml_href.split('_')[:-1])  # Remove the _MTL.txt

    mtl_metadata = MtlMetadata.from_file(mtl_xml_href, read_href_modifier)

    ang_href = ANG_ASSET_DEF.get_href(base_href)
    ang_metadata = AngMetadata.from_file(ang_href, read_href_modifier)

    scene_datetime = mtl_metadata.scene_datetime

    item = pystac.Item(id=mtl_metadata.scene_id,
                       bbox=mtl_metadata.bbox,
                       geometry=ang_metadata.get_scene_geometry(
                           mtl_metadata.bbox),
                       datetime=scene_datetime,
                       properties={})

    item.common_metadata.platform = L8_PLATFORM
    item.common_metadata.instruments = L8_INSTRUMENTS
    item.common_metadata.description = L8_ITEM_DESCRIPTION

    # eo
    item.ext.enable('eo')
    item.ext.eo.cloud_cover = mtl_metadata.cloud_cover

    # view
    item.ext.enable('view')
    item.ext.view.off_nadir = mtl_metadata.off_nadir
    item.ext.view.sun_elevation = mtl_metadata.sun_elevation
    # Sun Azimuth in landsat metadata is -180 to 180 from north, west being negative.
    # In STAC, it's 0 to 360 clockwise from north.
    sun_azimuth = mtl_metadata.sun_azimuth
    if sun_azimuth < 0.0:
        sun_azimuth = 360 + sun_azimuth
    item.ext.view.sun_azimuth = sun_azimuth

    # projection
    item.ext.enable('projection')
    item.ext.projection.epsg = mtl_metadata.epsg
    item.ext.projection.bbox = mtl_metadata.proj_bbox

    # landsat8
    item.stac_extensions.append(L8_EXTENSION_SCHEMA)
    item.properties.update(**mtl_metadata.additional_metadata)
    item.properties['landsat8:scene_id'] = ang_metadata.scene_id

    # -- Add assets

    # Add common assets
    for asset_definition in COMMON_ASSET_DEFS:
        asset_definition.add_asset(item, mtl_metadata, base_href)

    # Add SR assets
    for asset_definition in SR_ASSET_DEFS:
        asset_definition.add_asset(item, mtl_metadata, base_href)

    # Add thermal assets, if this is a L2SP product
    if mtl_metadata.processing_level == 'L2SP':
        for asset_definition in THERMAL_ASSET_DEFS:
            asset_definition.add_asset(item, mtl_metadata, base_href)

    # -- Add links

    usgs_item_page = (
        f"https://landsatlook.usgs.gov/stac-browser/collection02/level-2/standard/oli-tirs"
        f"/{scene_datetime.year}"
        f"/{mtl_metadata.wrs_path}/{mtl_metadata.wrs_row}"
        f"/{mtl_metadata.scene_id}")

    item.add_link(
        pystac.Link(rel="alternate",
                    target=usgs_item_page,
                    title="USGS stac-browser page",
                    media_type="text/html"))

    return item



from typing import List, Optional

import pystac
from pystac.extensions.eo import Band

from stactools.landsat.constants import (L8_SR_BANDS, L8_SP_BANDS)
from stactools.landsat.mtl_metadata import MtlMetadata


class AssetDef:
    """Holds asset defintions for L8 C2 Level 2 assets.

    If no explicit key is given, the key is the file suffix.
    """
    def __init__(self,
                 href_suffix: str,
                 media_type: str,
                 key: str,
                 title: Optional[str] = None,
                 description: Optional[str] = None,
                 bands: Optional[List[Band]] = None,
                 is_sr: bool = False,
                 is_thermal: bool = False,
                 is_qa: bool = False,
                 gsd: Optional[float] = None):
        self.href_suffix = href_suffix
        self.media_type = media_type
        self.key = key or href_suffix
        self.title = title
        self.description = description
        self.gsd = gsd
        self.bands = bands
        if bands is None:
            if is_sr:
                band = L8_SR_BANDS.get(href_suffix.split('.')[0])
                if band is not None:
                    self.bands = [band]
            elif is_thermal:
                band = L8_SP_BANDS.get(href_suffix.split('.')[0])
                if band is not None:
                    self.bands = [band]
        self.is_sr = is_sr
        self.is_thermal = is_thermal
        self.is_qa = is_qa

    def get_href(self, base_href: str) -> str:
        return f"{base_href}_{self.href_suffix}"

    def add_asset(self, item: pystac.Item, mtl_metadata: MtlMetadata,
                  base_href: str) -> None:
        asset = pystac.Asset(href=self.get_href(base_href),
                             media_type=self.media_type)
        if self.title:
            asset.title = self.title
        if self.description:
            asset.description = self.description

        # common_metadata

        if self.gsd is not None:
            item.common_metadata.set_gsd(self.gsd, asset)
        else:
            if self.is_sr or self.is_qa:
                sr_grd = mtl_metadata.sr_gsd
                if item.common_metadata.gsd != sr_grd:
                    item.common_metadata.set_gsd(sr_grd, asset)
            if self.is_thermal:
                thermal_grd = mtl_metadata.thermal_gsd
                if item.common_metadata.gsd != thermal_grd:
                    item.common_metadata.set_gsd(thermal_grd, asset)

        # eo

        if self.bands:
            asset.properties["eo:bands"] = [b.to_dict() for b in self.bands]

        # projection
        if self.is_sr or self.is_qa:
            item.ext.projection.set_shape(mtl_metadata.sr_shape, asset)
            item.ext.projection.set_transform(mtl_metadata.sr_transform, asset)
        if self.is_thermal:
            item.ext.projection.set_shape(mtl_metadata.thermal_shape, asset)
            item.ext.projection.set_transform(mtl_metadata.thermal_transform,
                                              asset)

        item.add_asset(self.key, asset)


THUMBNAIL_ASSET_DEF = AssetDef(key="thumbnail",
                               href_suffix="thumb_small.jpeg",
                               title="Thumbnail image",
                               media_type=pystac.MediaType.JPEG)
PREVIEW_ASSET_DEF = AssetDef(key="reduced_resolution_browse",
                             href_suffix="thumb_large.jpeg",
                             title="Reduced resolution browse image",
                             media_type=pystac.MediaType.JPEG)
ANG_ASSET_DEF = AssetDef(
    href_suffix="ANG.txt",
    key="ANG",
    title="Angle Coefficients File",
    description=("Collection 2 Level-1 Angle Coefficients File (ANG)"),
    media_type=pystac.MediaType.TEXT)
MTL_TXT_ASSET_DEF = AssetDef(
    href_suffix="MTL.txt",
    key="MTL.txt",
    title="Product Metadata File",
    description=("Collection 2 Level-1 Product Metadata File (MTL)"),
    media_type=pystac.MediaType.TEXT)
MTL_XML_ASSET_DEF = AssetDef(
    href_suffix="MTL.xml",
    key="MTL.xml",
    title="Product Metadata File (xml)",
    description="Collection 2 Level-1 Product Metadata File (xml)",
    media_type=pystac.MediaType.XML)
MTL_JSON_ASSET_DEF = AssetDef(
    href_suffix="MTL.json",
    key="MTL.json",
    title="Product Metadata File (json)",
    description="Collection 2 Level-1 Product Metadata File (json)",
    media_type=pystac.MediaType.JSON)
QA_PIXEL_ASSET_DEF = AssetDef(
    href_suffix="QA_PIXEL.TIF",
    key="QA_PIXEL",
    title="Pixel Quality Assessment Band",
    description="Collection 2 Level-1 Pixel Quality Assessment Band",
    media_type=pystac.MediaType.COG,
    is_qa=True)
QA_RADSAT_ASSET_DEF = AssetDef(
    href_suffix="QA_RADSAT.TIF",
    key="QA_RADSAT",
    title="Radiometric Saturation Quality Assessment Band",
    description=("Collection 2 Level-1 Radiometric Saturation "
                 "Quality Assessment Band"),
    media_type=pystac.MediaType.COG,
    is_qa=True)

COMMON_ASSET_DEFS = [
    THUMBNAIL_ASSET_DEF, PREVIEW_ASSET_DEF, ANG_ASSET_DEF, MTL_TXT_ASSET_DEF,
    MTL_XML_ASSET_DEF, MTL_JSON_ASSET_DEF, QA_PIXEL_ASSET_DEF,
    QA_RADSAT_ASSET_DEF
]

# Surface reflectance assets
# These assets are contained in both Level 2 Science Product (L2SP)
# and Level 2 Surface Reflectance (L2SR) processing levels
B1_ASSET_DEF = AssetDef(
    href_suffix="SR_B1.TIF",
    key="SR_B1",
    title="Coastal/Aerosol Band (B1)",
    description=("Collection 2 Level-2 Coastal/Aerosol Band "
                 "(B1) Surface Reflectance"),
    media_type=pystac.MediaType.COG,
    is_sr=True)
B2_ASSET_DEF = AssetDef(href_suffix="SR_B2.TIF",
                        key="SR_B2",
                        title="Blue Band (B2)",
                        description=("Collection 2 Level-2 Blue Band "
                                     "(B2) Surface Reflectance"),
                        media_type=pystac.MediaType.COG,
                        is_sr=True)
B3_ASSET_DEF = AssetDef(href_suffix="SR_B3.TIF",
                        key="SR_B3",
                        title="Green Band (B3)",
                        description=("Collection 2 Level-2 Green Band "
                                     "(B3) Surface Reflectance"),
                        media_type=pystac.MediaType.COG,
                        is_sr=True)
B4_ASSET_DEF = AssetDef(href_suffix="SR_B4.TIF",
                        key="SR_B4",
                        title="Red Band (B4)",
                        description=("Collection 2 Level-2 Red Band "
                                     "(B4) Surface Reflectance"),
                        media_type=pystac.MediaType.COG,
                        is_sr=True)
B5_ASSET_DEF = AssetDef(
    href_suffix="SR_B5.TIF",
    key="SR_B5",
    title="Near Infrared Band 0.8 (B5)",
    description=("Collection 2 Level-2 Near Infrared Band 0.8 "
                 "(B5) Surface Reflectance"),
    media_type=pystac.MediaType.COG,
    is_sr=True)
B6_ASSET_DEF = AssetDef(
    href_suffix="SR_B6.TIF",
    key="SR_B6",
    title="Short-wave Infrared Band 1.6 (B6)",
    description=("Collection 2 Level-2 Short-wave Infrared Band 1.6 "
                 "(B6) Surface Reflectance"),
    media_type=pystac.MediaType.COG,
    is_sr=True)
B7_ASSET_DEF = AssetDef(
    href_suffix="SR_B7.TIF",
    key="SR_B7",
    title="Short-wave Infrared Band 2.2 (B7)",
    description=("Collection 2 Level-2 Short-wave Infrared Band 2.2 "
                 "(B7) Surface Reflectance"),
    media_type=pystac.MediaType.COG,
    is_sr=True)
QA_AEROSOL_ASSET_DEF = AssetDef(
    href_suffix="SR_QA_AEROSOL.TIF",
    key="SR_B8",
    title="Aerosol Quality Analysis Band",
    description=("Collection 2 Level-2 Aerosol Quality Analysis Band "
                 "(ANG) Surface Reflectance"),
    media_type=pystac.MediaType.COG,
    is_qa=True)

SR_ASSET_DEFS = [
    B1_ASSET_DEF, B2_ASSET_DEF, B3_ASSET_DEF, B4_ASSET_DEF, B5_ASSET_DEF,
    B6_ASSET_DEF, B7_ASSET_DEF, QA_AEROSOL_ASSET_DEF
]

# Thermal assets
# These assets are contained only in the Level 2 Science Product (L2SP)
# processing level
ST_B10_ASSET_DEF = AssetDef(
    # COG has Pixel Size of 30m,
    # but 100m spatial resolution at the sensor
    gsd=100.0,
    href_suffix="ST_B10.TIF",
    key="ST_B10",
    title="Surface Temperature Band (B10)",
    description=("Landsat Collection 2 Level-2 Surface Temperature Band "
                 "(B10) Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_ATRAN_ASSET_DEF = AssetDef(
    href_suffix="ST_ATRAN.TIF",
    key="ST_ATRAN",
    title="Atmospheric Transmittance Band",
    description=("Landsat Collection 2 Level-2 Atmospheric "
                 "Transmittance Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_CDIST_ASSET_DEF = AssetDef(
    href_suffix="ST_CDIST.TIF",
    key="ST_CDIST",
    title="Cloud Distance Band",
    description=("Landsat Collection 2 Level-2 Cloud Distance Band "
                 "Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_DRAD_ASSET_DEF = AssetDef(
    href_suffix="ST_DRAD.tif",
    key="ST_DRAD",
    title="Downwelled Radiance Band",
    description=("Landsat Collection 2 Level-2 Downwelled "
                 "Radiance Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_URAD_ASSET_DEF = AssetDef(
    href_suffix="ST_URAD.TIF",
    key="ST_URAD",
    title="Upwelled Radiance Band",
    description=("Landsat Collection 2 Level-2 Upwelled "
                 "Radiance Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_TRAD_ASSET_DEF = AssetDef(
    href_suffix="ST_TRAD.TIF",
    key="ST_TRAD",
    title="Thermal Radiance Band",
    description=("Landsat Collection 2 Level-2 Thermal Radiance "
                 "Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_EMIS_ASSET_DEF = AssetDef(
    href_suffix="ST_EMIS.TIF",
    key="ST_EMIS",
    title="Emissivity Band",
    description=("Landsat Collection 2 Level-2 Emissivity Band "
                 "Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_EMSD_ASSET_DEF = AssetDef(
    href_suffix="ST_EMSD.TIF",
    key="ST_EMSD",
    title="Emissivity Standard Deviation Band",
    description=("Landsat Collection 2 Level-2 Emissivity "
                 "Standard Deviation Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

ST_QA_ASSET_DEF = AssetDef(
    href_suffix="ST_QA.TIF",
    key="ST_QA",
    title="Surface Temperature Quality Assessment Band",
    description=("Landsat Collection 2 Level-2 Surface Temperature "
                 "Band Surface Temperature Product"),
    media_type=pystac.MediaType.COG,
    is_thermal=True)

THERMAL_ASSET_DEFS = [
    ST_B10_ASSET_DEF, ST_ATRAN_ASSET_DEF, ST_CDIST_ASSET_DEF,
    ST_DRAD_ASSET_DEF, ST_URAD_ASSET_DEF, ST_TRAD_ASSET_DEF, ST_EMIS_ASSET_DEF,
    ST_EMSD_ASSET_DEF, ST_QA_ASSET_DEF
]



from datetime import datetime
from typing import Any, Dict, List, Optional

from pystac.utils import str_to_datetime

from stactools.core.utils import map_opt
from stactools.core.projection import transform_from_bbox
from stactools.core.io import ReadHrefModifier
from stactools.core.io.xml import XmlElement


class MTLError(Exception):
    pass


class MtlMetadata:
    """Parses a Collection 2 MTL XML file.

    References https://github.com/sat-utils/sat-stac-landsat/blob/f2263485043a827b4153aecc12f45a3d1363e9e2/satstac/landsat/main.py#L157
    """  # noqa

    def __init__(self, root: XmlElement, href: Optional[str] = None):
        self._root = root
        self.href = href

    def _xml_error(self, item: str) -> MTLError:
        return MTLError(f"Cannot find {item} in MTL metadata" +
                        ("" if self.href is None else f" at {self.href}"))

    def _get_text(self, xpath: str) -> str:
        return self._root.find_text_or_throw(xpath, self._xml_error)

    def _get_float(self, xpath: str) -> float:
        return float(self._get_text(xpath))

    def _get_int(self, xpath: str) -> int:
        return int(self._get_text(xpath))

    @property
    def scene_id(self) -> str:
        return self._get_text("PRODUCT_CONTENTS/LANDSAT_PRODUCT_ID")

    @property
    def processing_level(self) -> str:
        """Processing level. Determines product contents.

        Returns either 'L2SP' or 'L2SR', standing for
        'Level 2 Science Product' and 'Level 2 Surface Reflectance',
        respectively. L2SP has thermal + surface reflectance assets;
        L2SR only has surface reflectance.
        """
        return self._get_text("PRODUCT_CONTENTS/PROCESSING_LEVEL")

    @property
    def epsg(self) -> int:
        utm_zone = self._root.find_text('PROJECTION_ATTRIBUTES/UTM_ZONE')
        if utm_zone:
            bbox = self.bbox
            utm_zone = self._get_text('PROJECTION_ATTRIBUTES/UTM_ZONE')
            center_lat = (bbox[1] + bbox[3]) / 2.0
            return int(f"{326 if center_lat > 0 else 327}{utm_zone}")
        else:
            # Polar Stereographic
            # Based on Landsat 8-9 OLI/TIRS Collection 2 Level 1 Data Format Control Book,
            # should only ever be 71 or -71
            lat_ts = self._get_text('PROJECTION_ATTRIBUTES/TRUE_SCALE_LAT')
            if lat_ts == "-71.00000":
                # Antarctic
                return 3031
            elif lat_ts == "71.00000":
                # Arctic
                return 3995
            else:
                raise MTLError(
                    f'Unexpeced value for PROJECTION_ATTRIBUTES/TRUE_SCALE_LAT: {lat_ts} '
                )

    @property
    def bbox(self) -> List[float]:
        lons = [
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_UL_LON_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_UR_LON_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_LL_LON_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_LR_LON_PRODUCT")
        ]

        lats = [
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_UL_LAT_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_UR_LAT_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_LL_LAT_PRODUCT"),
            self._get_float("PROJECTION_ATTRIBUTES/CORNER_LR_LAT_PRODUCT")
        ]

        return [min(lons), min(lats), max(lons), max(lats)]

    @property
    def proj_bbox(self) -> List[float]:
        xs = [
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_UL_PROJECTION_X_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_UR_PROJECTION_X_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_LL_PROJECTION_X_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_LR_PROJECTION_X_PRODUCT")
        ]

        ys = [
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_UL_PROJECTION_Y_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_UR_PROJECTION_Y_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_LL_PROJECTION_Y_PRODUCT"),
            self._get_float(
                "PROJECTION_ATTRIBUTES/CORNER_LR_PROJECTION_Y_PRODUCT")
        ]

        return [min(xs), min(ys), max(xs), max(ys)]

    @property
    def sr_shape(self) -> List[int]:
        """Shape for surface reflectance assets.

        Used for proj:shape. In [row, col] order"""
        return [
            self._get_int("PROJECTION_ATTRIBUTES/REFLECTIVE_LINES"),
            self._get_int("PROJECTION_ATTRIBUTES/REFLECTIVE_SAMPLES")
        ]

    @property
    def thermal_shape(self) -> Optional[List[int]]:
        """Shape for thermal bands (Bands 10–11).

        None if thermal bands not present.
        Used for proj:shape. In [row, col] order"""
        rows = map_opt(
            int, self._root.find_text("PROJECTION_ATTRIBUTES/THERMAL_LINES"))
        cols = map_opt(
            int, self._root.find_text("PROJECTION_ATTRIBUTES/THERMAL_SAMPLES"))

        if rows is not None and cols is not None:
            return [rows, cols]
        else:
            return None

    @property
    def sr_transform(self) -> List[float]:
        return transform_from_bbox(self.proj_bbox, self.sr_shape)

    @property
    def thermal_transform(self) -> Optional[List[float]]:
        return map_opt(
            lambda shape: transform_from_bbox(self.proj_bbox, shape),
            self.thermal_shape)

    @property
    def sr_gsd(self) -> float:
        return self._get_float(
            "LEVEL1_PROJECTION_PARAMETERS/GRID_CELL_SIZE_THERMAL")

    @property
    def thermal_gsd(self) -> Optional[float]:
        return map_opt(
            float,
            self._root.find_text(
                'LEVEL1_PROJECTION_PARAMETERS/GRID_CELL_SIZE_THERMAL'))

    @property
    def scene_datetime(self) -> datetime:
        date = self._get_text("IMAGE_ATTRIBUTES/DATE_ACQUIRED")
        time = self._get_text("IMAGE_ATTRIBUTES/SCENE_CENTER_TIME")

        return str_to_datetime(f"{date} {time}")

    @property
    def cloud_cover(self) -> float:
        return self._get_float("IMAGE_ATTRIBUTES/CLOUD_COVER")

    @property
    def sun_azimuth(self) -> float:
        return self._get_float("IMAGE_ATTRIBUTES/SUN_AZIMUTH")

    @property
    def sun_elevation(self) -> float:
        return self._get_float("IMAGE_ATTRIBUTES/SUN_ELEVATION")

    @property
    def off_nadir(self) -> Optional[float]:
        if self._get_text("IMAGE_ATTRIBUTES/NADIR_OFFNADIR") == "NADIR":
            return 0
        else:
            return None

    @property
    def wrs_path(self) -> str:
        return self._get_text("IMAGE_ATTRIBUTES/WRS_PATH").zfill(3)

    @property
    def wrs_row(self) -> str:
        return self._get_text("IMAGE_ATTRIBUTES/WRS_ROW").zfill(3)

    @property
    def additional_metadata(self) -> Dict[str, Any]:
        return {
            "landsat:cloud_cover_land":
            self._get_float("IMAGE_ATTRIBUTES/CLOUD_COVER_LAND"),
            "landsat:wrs_type":
            self._get_text("IMAGE_ATTRIBUTES/WRS_TYPE"),
            "landsat:wrs_path":
            self.wrs_path,
            "landsat:wrs_row":
            self.wrs_row,
            "landsat:collection_category":
            self._get_text("PRODUCT_CONTENTS/COLLECTION_CATEGORY"),
            "landsat:collection_number":
            self._get_text("PRODUCT_CONTENTS/COLLECTION_NUMBER"),
            "landsat:processing_level":
            self.processing_level
        }

    @classmethod
    def from_file(cls,
                  href,
                  read_href_modifier: Optional[ReadHrefModifier] = None
                  ) -> "MtlMetadata":
        return cls(XmlElement.from_file(href, read_href_modifier), href=href)



import os
from imp import load_source
from setuptools import setup, find_namespace_packages
from glob import glob
import io

name = "stactools_landsat"
description = ("Subpackage for working with landsat data of stactools, "
               "a command line tool and Python library for working with STAC.")

__version__ = load_source(
    "stactools.landsat.version",
    os.path.join(os.path.dirname(__file__), "stactools/landsat/version.py"),
).__version__

from os.path import basename, splitext

here = os.path.abspath(os.path.dirname(__file__))

# get the dependencies and installs
with io.open(os.path.join(here, "requirements.txt"), encoding="utf-8") as f:
    install_requires = f.read().split("\n")

# Add stactools subpackage dependencies
install_requires.extend(["stactools_core=={}".format(__version__)])

with open(os.path.join(here, "README.md")) as readme_file:
    readme = readme_file.read()

setup(
    name=name,
    description=description,
    version=__version__,
    long_description=readme,
    long_description_content_type="text/markdown",
    author="stac-utils",
    author_email="stac@radiant.earth",
    url="https://github.com/stac-utils/stactools.git",
    packages=find_namespace_packages(),
    py_modules=[
        splitext(basename(path))[0] for path in glob("stactools/*.py")
    ],
    include_package_data=False,
    install_requires=install_requires,
    license="Apache Software License 2.0",
    keywords=["stactools", "psytac", "imagery", "raster", "catalog", "STAC"],
)



import logging
from typing import Any, Dict, List, Optional

from stactools.core.io import ReadHrefModifier, read_text

logger = logging.getLogger(__name__)


class AngError(Exception):
    pass


class AngMetadata:
    """Represents the Angle Coefficient file (ANG) file.

    The ANG file can be used to derive the scene geometry.

    Modified from https://github.com/sat-utils/sat-stac-landsat/blob/485eb6874cf0e20744a19970c6d8a295eee5942d/satstac/landsat/main.py#L108
    """  # noqa

    def __init__(self, ang_txt: str, href: Optional[str] = None):
        self._sz = []
        self._coords = []
        scene_id = None

        try:
            for line in ang_txt.split('\n'):
                if 'LANDSAT_SCENE_ID' in line:
                    scene_id = line.split('=')[1].strip(' "')
                if 'BAND01_NUM_L1T_LINES' in line or 'BAND01_NUM_L1T_SAMPS' in line:
                    self._sz.append(float(line.split('=')[1]))
                if ('BAND01_L1T_IMAGE_CORNER_LINES' in line
                        or 'BAND01_L1T_IMAGE_CORNER_SAMPS' in line):
                    self._coords.append([
                        float(x) for x in line.split('=')[1].strip().strip(
                            '()').split(',')
                    ])
                if len(self._coords) == 2:
                    break
        except Exception as e:
            href_msg = "" if href is None else f" from {href}"
            raise AngError(f"Could not parse ANG file{href_msg}") from e

        if scene_id is None:
            href_msg = "" if href is None else f" from {href}"
            raise AngError(f"Could not get scene id {href_msg}")

        self.scene_id = scene_id

    def get_scene_geometry(self, bbox: List[float]) -> Dict[str, Any]:
        """Get the scene geometry from the ANG file based on the
        bbox supplied by the MTL.

        TODO: This produces a footprint that is larger than the ones provided
        by the USGS STAC Items. Determine how to more closely match USGS, or
        get them to publish the footprint geom as an additional product file.
        """
        dlon = bbox[2] - bbox[0]
        dlat = bbox[3] - bbox[1]
        lons = [c / self._sz[1] * dlon + bbox[0] for c in self._coords[1]]
        lats = [((self._sz[0] - c) / self._sz[0]) * dlat + bbox[1]
                for c in self._coords[0]]
        coordinates = [[[lons[0], lats[0]], [lons[1], lats[1]],
                        [lons[2], lats[2]], [lons[3], lats[3]],
                        [lons[0], lats[0]]]]
        return {'type': 'Polygon', 'coordinates': coordinates}

    @classmethod
    def from_file(cls,
                  href,
                  read_href_modifier: Optional[ReadHrefModifier] = None
                  ) -> "AngMetadata":
        return cls(read_text(href, read_href_modifier), href=href)
